==> dkist_processing_dlnirsp-0.1.0rc3.tar.gz: dkist_processing_dlnirsp-0.1.0rc3/dkist_processing_dlnirsp/tests/test_science.py <==
import json

import numpy as np
import pytest
from astropy.io import fits
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.task_name import TaskName
from dkist_processing_common.tests.conftest import FakeGQLClient

from dkist_processing_dlnirsp.models.tags import DlnirspTag
from dkist_processing_dlnirsp.tasks.science import ScienceCalibration
from dkist_processing_dlnirsp.tests.conftest import DlnirspTestingConstants
from dkist_processing_dlnirsp.tests.conftest import DlnirspTestingParameters
from dkist_processing_dlnirsp.tests.conftest import SimpleModulatedHeaders
from dkist_processing_dlnirsp.tests.conftest import tag_obs_on_mosaic_and_modstate
from dkist_processing_dlnirsp.tests.conftest import write_dark_frames_to_task
from dkist_processing_dlnirsp.tests.conftest import write_geometric_calibration_to_task
from dkist_processing_dlnirsp.tests.conftest import write_observe_frames_to_task
from dkist_processing_dlnirsp.tests.conftest import write_simple_frames_to_task
from dkist_processing_dlnirsp.tests.conftest import write_solar_gain_frames_to_task


@pytest.fixture
def dark_signal() -> float:
    return 100.0


@pytest.fixture
def solar_signal() -> float:
    return 2000.0


@pytest.fixture
def true_stokes_science_signal() -> np.ndarray:
    return np.array([4000.0, -1000.0, 2000.0, 1000.0])


@pytest.fixture
def modulated_science_signal(true_stokes_science_signal, modulation_matrix) -> np.ndarray:
    modulated_data = modulation_matrix @ true_stokes_science_signal  # shape = (8,)
    return modulated_data


@pytest.fixture
def make_dark_data(dark_signal):
    def make_array(frame: Spec122Dataset):
        shape = frame.array_shape[1:]
        return np.ones(shape) * dark_signal

    return make_array


@pytest.fixture
def make_solar_data(solar_signal):
    def make_array(frame: Spec122Dataset):
        shape = frame.array_shape[1:]
        return np.ones(shape) * solar_signal

    return make_array


@pytest.fixture
def make_full_demodulation_matrix(demodulation_matrix):
    def make_array(frame: Spec122Dataset):
        array_shape = frame.array_shape[1:]
        return np.ones(array_shape + demodulation_matrix.shape) * demodulation_matrix

    return make_array


@pytest.fixture
def make_linearized_science_data(
    dark_signal, solar_signal, modulated_science_signal, modulation_matrix
):
    def make_array(frame: SimpleModulatedHeaders):
        shape = frame.array_shape[1:]
        modstate = frame.current_modstate("foo") - 1
        raw_data = np.ones(shape) * modulated_science_signal[modstate]
        obs_data = (raw_data * solar_signal) + dark_signal
        return obs_data

    return make_array


@pytest.fixture
def science_task_with_data(
    tmp_path,
    recipe_run_id,
    link_constants_db,
    assign_input_dataset_doc_to_task,
    constants_class_with_different_num_slits,
    is_polarimetric,
    jband_group_id_file_parameter,
    jband_ifu_x_pos_file_parameter,
    jband_ifu_y_pos_file_parameter,
):
    science_obs_time = 1.0
    num_modstates = 8 if is_polarimetric else 1
    num_X_tiles = 2
    num_Y_tiles = 3
    num_mosaic = 4
    pol_mode = "Full Stokes" if is_polarimetric else "Stokes I"
    constants = DlnirspTestingConstants(
        OBSERVE_EXPOSURE_TIMES=(science_obs_time,),
        NUM_MODSTATES=num_modstates,
        NUM_SPATIAL_STEPS_X=num_X_tiles,
        NUM_SPATIAL_STEPS_Y=num_Y_tiles,
        NUM_MOSAIC_REPEATS=num_mosaic,
        POLARIMETER_MODE=pol_mode,
    )
    link_constants_db(
        recipe_run_id=recipe_run_id,
        constants_obj=constants,
    )

    with ScienceCalibration(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        parameters = DlnirspTestingParameters(
            dlnirsp_group_id_file_jband=jband_group_id_file_parameter,
            dlnirsp_ifu_x_pos_file_jband=jband_ifu_x_pos_file_parameter,
            dlnirsp_ifu_y_pos_file_jband=jband_ifu_y_pos_file_parameter,
        )
        assign_input_dataset_doc_to_task(
            task,
            parameters,
        )

        task.scratch = WorkflowFileSystem(scratch_base_path=tmp_path, recipe_run_id=recipe_run_id)
        task.constants = constants_class_with_different_num_slits(
            recipe_run_id=recipe_run_id, task_name="test"
        )

        yield task, science_obs_time, num_modstates, num_X_tiles, num_Y_tiles, num_mosaic


@pytest.mark.parametrize(
    "is_polarimetric",
    [pytest.param(True, id="polarimetric"), pytest.param(False, id="spectrographic")],
)
def test_science_task_completes(
    science_task_with_data,
    is_polarimetric,
    make_dark_data,
    make_solar_data,
    shifts_and_scales,
    reference_wave_axis,
    make_full_demodulation_matrix,
    make_linearized_science_data,
    jband_group_id_array,
    mocker,
):
    """
    Given: A ScienceTask with all intermediate calibrations and a set of linearized OBSERVE frames
    When: Running the task
    Then: The task completes and the expected number of files are produced

    NOTE: We don't really check anything about correctness in this test. That's a GROGU thing.
    """
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    (
        task,
        science_obs_time,
        num_modstates,
        num_X_tiles,
        num_Y_tiles,
        num_mosaic,
    ) = science_task_with_data
    array_shape = jband_group_id_array.shape

    write_dark_frames_to_task(
        task,
        array_shape=array_shape,
        exp_time_ms=science_obs_time,
        tags=[
            DlnirspTag.intermediate(),
            DlnirspTag.task("DARK"),
            DlnirspTag.exposure_time(science_obs_time),
        ],
        data_func=make_dark_data,
    )

    write_solar_gain_frames_to_task(
        task,
        array_shape=array_shape,
        num_modstates=1,
        tags=[
            DlnirspTag.intermediate(),
            DlnirspTag.task("SOLAR_GAIN"),
        ],
        data_func=make_solar_data,
    )

    shift_dict, scale_dict, _, _ = shifts_and_scales
    write_geometric_calibration_to_task(
        task, shift_dict=shift_dict, scale_dict=scale_dict, wave_axis=reference_wave_axis
    )

    if is_polarimetric:
        write_simple_frames_to_task(
            task,
            task_type="POLCAL",
            array_shape=array_shape,
            num_modstates=1,
            tags=[DlnirspTag.intermediate(), DlnirspTag.task("DEMOD_MATRICES")],
            data_func=make_full_demodulation_matrix,
        )

    num_observe_frames = write_observe_frames_to_task(
        task,
        exp_time_ms=science_obs_time,
        array_shape=array_shape,
        num_modstates=num_modstates,
        num_X_tiles=num_X_tiles,
        num_Y_tiles=num_Y_tiles,
        num_mosaics=num_mosaic,
        tags=[DlnirspTag.linearized(), DlnirspTag.task("OBSERVE")],
        data_func=make_linearized_science_data,
        tag_func=tag_obs_on_mosaic_and_modstate,
    )

    task()

    # TODO: Uncomment once FYPOS NaN issue is solved
    # Test that the (currently crucial) debug frames were written
    # for name in ["ID", "FXPOS", "FYPOS"]:
    #     tags = [DlnirspTag.debug(), DlnirspTag.task(f"REMAPPED_IFU_{name}")]
    #     file_list = list(task.read(tags=tags))
    #     assert len(file_list) == 1
    #     assert file_list[0].exists

    for mosaic in range(num_mosaic):
        for X_tile in range(num_X_tiles):
            for Y_tile in range(num_Y_tiles):
                for stokes in task.constants.stokes_params:
                    tags = [
                        DlnirspTag.calibrated(),
                        DlnirspTag.frame(),
                        DlnirspTag.mosaic_num(mosaic),
                        DlnirspTag.tile_X_num(X_tile),
                        DlnirspTag.tile_Y_num(Y_tile),
                        DlnirspTag.stokes(stokes),
                    ]
                    file_list = list(task.read(tags=tags))
                    if not is_polarimetric and stokes in ["Q", "U", "V"]:
                        assert len(file_list) == 0
                    else:
                        assert len(file_list) == 1
                        assert file_list[0].exists

                        header = fits.getheader(file_list[0])
                        assert "DATE-END" in header
                        if is_polarimetric:
                            assert "POL_NOIS" in header
                            assert "POL_SENS" in header
                        else:
                            assert "POL_NOIS" not in header
                            assert "POL_SENS" not in header

    quality_files = list(task.read(tags=[DlnirspTag.quality("TASK_TYPES")]))
    assert len(quality_files) == 1
    file = quality_files[0]
    with file.open() as f:
        data = json.load(f)
        assert isinstance(data, dict)
        assert data["task_type"] == TaskName.observe.value
        assert data["total_frames"] == num_observe_frames
        assert data["frames_not_used"] == 0

==> dkist_processing_dlnirsp-0.2.1.tar.gz: dkist_processing_dlnirsp-0.2.1/dkist_processing_dlnirsp/tests/test_science.py <==
import json

import numpy as np
import pytest
from astropy.io import fits
from dkist_data_simulator.spec122 import Spec122Dataset
from dkist_processing_common._util.scratch import WorkflowFileSystem
from dkist_processing_common.models.task_name import TaskName
from dkist_processing_common.tests.conftest import FakeGQLClient

from dkist_processing_dlnirsp.models.tags import DlnirspTag
from dkist_processing_dlnirsp.tasks.science import ScienceCalibration
from dkist_processing_dlnirsp.tests.conftest import DlnirspTestingConstants
from dkist_processing_dlnirsp.tests.conftest import DlnirspTestingParameters
from dkist_processing_dlnirsp.tests.conftest import SimpleModulatedHeaders
from dkist_processing_dlnirsp.tests.conftest import tag_obs_on_mosaic_and_modstate
from dkist_processing_dlnirsp.tests.conftest import write_dark_frames_to_task
from dkist_processing_dlnirsp.tests.conftest import write_geometric_calibration_to_task
from dkist_processing_dlnirsp.tests.conftest import write_observe_frames_to_task
from dkist_processing_dlnirsp.tests.conftest import write_simple_frames_to_task
from dkist_processing_dlnirsp.tests.conftest import write_solar_gain_frames_to_task


@pytest.fixture
def dark_signal() -> float:
    return 100.0


@pytest.fixture
def solar_signal() -> float:
    return 2000.0


@pytest.fixture
def true_stokes_science_signal() -> np.ndarray:
    return np.array([4000.0, -1000.0, 2000.0, 1000.0])


@pytest.fixture
def modulated_science_signal(true_stokes_science_signal, modulation_matrix) -> np.ndarray:
    modulated_data = modulation_matrix @ true_stokes_science_signal  # shape = (8,)
    return modulated_data


@pytest.fixture
def make_dark_data(dark_signal):
    def make_array(frame: Spec122Dataset):
        shape = frame.array_shape[1:]
        return np.ones(shape) * dark_signal

    return make_array


@pytest.fixture
def make_solar_data(solar_signal):
    def make_array(frame: Spec122Dataset):
        shape = frame.array_shape[1:]
        return np.ones(shape) * solar_signal

    return make_array


@pytest.fixture
def make_full_demodulation_matrix(demodulation_matrix):
    def make_array(frame: Spec122Dataset):
        array_shape = frame.array_shape[1:]
        return np.ones(array_shape + demodulation_matrix.shape) * demodulation_matrix

    return make_array


@pytest.fixture
def make_linearized_science_data(
    dark_signal, solar_signal, modulated_science_signal, modulation_matrix
):
    def make_array(frame: SimpleModulatedHeaders):
        shape = frame.array_shape[1:]
        modstate = frame.current_modstate("foo") - 1
        raw_data = np.ones(shape) * modulated_science_signal[modstate]
        obs_data = (raw_data * solar_signal) + dark_signal
        return obs_data

    return make_array


@pytest.fixture
def science_task_with_data(
    tmp_path,
    recipe_run_id,
    link_constants_db,
    assign_input_dataset_doc_to_task,
    constants_class_with_different_num_slits,
    is_polarimetric,
    jband_group_id_file_parameter,
    jband_ifu_x_pos_file_parameter,
    jband_ifu_y_pos_file_parameter,
):
    science_obs_time = 1.0
    num_modstates = 8 if is_polarimetric else 1
    num_X_tiles = 2
    num_Y_tiles = 3
    num_mosaic = 4
    pol_mode = "Full Stokes" if is_polarimetric else "Stokes I"
    constants = DlnirspTestingConstants(
        OBSERVE_EXPOSURE_TIMES=(science_obs_time,),
        NUM_MODSTATES=num_modstates,
        NUM_SPATIAL_STEPS_X=num_X_tiles,
        NUM_SPATIAL_STEPS_Y=num_Y_tiles,
        NUM_MOSAIC_REPEATS=num_mosaic,
        POLARIMETER_MODE=pol_mode,
    )
    link_constants_db(
        recipe_run_id=recipe_run_id,
        constants_obj=constants,
    )

    with ScienceCalibration(
        recipe_run_id=recipe_run_id,
        workflow_name="workflow_name",
        workflow_version="workflow_version",
    ) as task:
        parameters = DlnirspTestingParameters(
            dlnirsp_group_id_file_jband=jband_group_id_file_parameter,
            dlnirsp_ifu_x_pos_file_jband=jband_ifu_x_pos_file_parameter,
            dlnirsp_ifu_y_pos_file_jband=jband_ifu_y_pos_file_parameter,
        )
        assign_input_dataset_doc_to_task(
            task,
            parameters,
        )

        task.scratch = WorkflowFileSystem(scratch_base_path=tmp_path, recipe_run_id=recipe_run_id)
        task.constants = constants_class_with_different_num_slits(
            recipe_run_id=recipe_run_id, task_name="test"
        )

        yield task, science_obs_time, num_modstates, num_X_tiles, num_Y_tiles, num_mosaic


@pytest.mark.parametrize(
    "is_polarimetric",
    [pytest.param(True, id="polarimetric"), pytest.param(False, id="spectrographic")],
)
def test_science_task_completes(
    science_task_with_data,
    is_polarimetric,
    make_dark_data,
    make_solar_data,
    shifts_and_scales,
    reference_wave_axis,
    make_full_demodulation_matrix,
    make_linearized_science_data,
    jband_group_id_array,
    mocker,
):
    """
    Given: A ScienceTask with all intermediate calibrations and a set of linearized OBSERVE frames
    When: Running the task
    Then: The task completes and the expected number of files are produced

    NOTE: We don't really check anything about correctness in this test. That's a GROGU thing.
    """
    mocker.patch(
        "dkist_processing_common.tasks.mixin.metadata_store.GraphQLClient", new=FakeGQLClient
    )
    (
        task,
        science_obs_time,
        num_modstates,
        num_X_tiles,
        num_Y_tiles,
        num_mosaic,
    ) = science_task_with_data
    array_shape = jband_group_id_array.shape

    write_dark_frames_to_task(
        task,
        array_shape=array_shape,
        exp_time_ms=science_obs_time,
        tags=[
            DlnirspTag.intermediate(),
            DlnirspTag.task_dark(),
            DlnirspTag.exposure_time(science_obs_time),
        ],
        data_func=make_dark_data,
    )

    write_solar_gain_frames_to_task(
        task,
        array_shape=array_shape,
        num_modstates=1,
        tags=[
            DlnirspTag.intermediate(),
            DlnirspTag.task_solar_gain(),
        ],
        data_func=make_solar_data,
    )

    shift_dict, scale_dict, _, _ = shifts_and_scales
    write_geometric_calibration_to_task(
        task, shift_dict=shift_dict, scale_dict=scale_dict, wave_axis=reference_wave_axis
    )

    if is_polarimetric:
        write_simple_frames_to_task(
            task,
            task_type=TaskName.polcal.value,
            array_shape=array_shape,
            num_modstates=1,
            tags=[DlnirspTag.intermediate(), DlnirspTag.task_demodulation_matrices()],
            data_func=make_full_demodulation_matrix,
        )

    num_observe_frames = write_observe_frames_to_task(
        task,
        exp_time_ms=science_obs_time,
        array_shape=array_shape,
        num_modstates=num_modstates,
        num_X_tiles=num_X_tiles,
        num_Y_tiles=num_Y_tiles,
        num_mosaics=num_mosaic,
        tags=[DlnirspTag.linearized(), DlnirspTag.task_observe()],
        data_func=make_linearized_science_data,
        tag_func=tag_obs_on_mosaic_and_modstate,
    )

    task()

    # TODO: Uncomment once FYPOS NaN issue is solved
    # Test that the (currently crucial) debug frames were written
    # for name in ["ID", "FXPOS", "FYPOS"]:
    #     tags = [DlnirspTag.debug(), DlnirspTag.task(f"REMAPPED_IFU_{name}")]
    #     file_list = list(task.read(tags=tags))
    #     assert len(file_list) == 1
    #     assert file_list[0].exists

    for mosaic in range(num_mosaic):
        for X_tile in range(num_X_tiles):
            for Y_tile in range(num_Y_tiles):
                for stokes in task.constants.stokes_params:
                    tags = [
                        DlnirspTag.calibrated(),
                        DlnirspTag.frame(),
                        DlnirspTag.mosaic_num(mosaic),
                        DlnirspTag.tile_X_num(X_tile),
                        DlnirspTag.tile_Y_num(Y_tile),
                        DlnirspTag.stokes(stokes),
                    ]
                    file_list = list(task.read(tags=tags))
                    if not is_polarimetric and stokes in ["Q", "U", "V"]:
                        assert len(file_list) == 0
                    else:
                        assert len(file_list) == 1
                        assert file_list[0].exists

                        header = fits.getheader(file_list[0])
                        assert "DATE-END" in header
                        if is_polarimetric:
                            assert "POL_NOIS" in header
                            assert "POL_SENS" in header
                        else:
                            assert "POL_NOIS" not in header
                            assert "POL_SENS" not in header

    quality_files = list(task.read(tags=[DlnirspTag.quality("TASK_TYPES")]))
    assert len(quality_files) == 1
    file = quality_files[0]
    with file.open() as f:
        data = json.load(f)
        assert isinstance(data, dict)
        assert data["task_type"] == TaskName.observe.value
        assert data["total_frames"] == num_observe_frames
        assert data["frames_not_used"] == 0
